"""Module for _PhraseMatcher: base class for other phrase based spaczz matchers."""
from __future__ import annotations

from collections import defaultdict
from typing import (
    Any,
    Callable,
    DefaultDict,
    Generator,
    Iterable,
    Optional,
)
import warnings

from spacy.tokens import Doc
from spacy.vocab import Vocab

from ..exceptions import KwargsWarning
from ..search import _PhraseSearcher


class _PhraseMatcher:
    """spaCy-like matcher for finding flexible matches in `Doc` objects.

    Matches added patterns against the `Doc` object it is called on.
    Accepts labeled patterns in the form of `Doc` objects.

    Attributes:
        defaults: Keyword arguments to be used as default matching settings.
            See `_PhraseSearcher` documentation for details.
        name: Class attribute - the name of the matcher.
        type: The kind of matcher object.
        _callbacks:
            On match functions to modify `Doc` objects passed to the matcher.
            Can make use of the matches identified.
        _patterns:
            Patterns added to the matcher. Contains patterns
            and kwargs that should be used during matching
            for each labels added.
    """

    name = "_phrase_matcher"

    def __init__(self: _PhraseMatcher, vocab: Vocab, **defaults: Any) -> None:
        """Initializes the base phrase matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are currently pure
                Python and do not share vocabulary
                with spaCy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings.
                These arguments will become the new defaults for matching.
                See `_PhraseSearcher` documentation for details.
        """
        self.defaults = defaults
        self.type = "_phrase"
        self._callbacks: dict[
            str,
            Optional[Callable[[Any, Doc, int, list[tuple[str, int, int, int]]], None]],
        ] = {}  # Any type due to inheritence typing issue.
        self._patterns: DefaultDict[str, DefaultDict[str, Any]] = defaultdict(
            lambda: defaultdict(list)
        )
        self._searcher = _PhraseSearcher(vocab=vocab)

    def __call__(self: _PhraseMatcher, doc: Doc) -> list[tuple[str, int, int, int]]:
        """Find all sequences matching the supplied patterns in the doc.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of (key, start, end, ratio) tuples, describing the matches.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> doc = nlp("Ridley Scott was the director of Alien.")
            >>> matcher.add("NAME", [nlp("Ridley Scott")])
            >>> matcher(doc)
            [('NAME', 0, 2, 100)]
        """
        matches = set()
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(patterns["patterns"], patterns["kwargs"]):
                if not kwargs:
                    kwargs = self.defaults
                matches_wo_label = self._searcher.match(doc, pattern, **kwargs)
                if matches_wo_label:
                    matches_w_label = [
                        (label,) + match_wo_label for match_wo_label in matches_wo_label
                    ]
                    for match in matches_w_label:
                        matches.add(match)
        if matches:
            sorted_matches = sorted(matches, key=lambda x: (x[1], -x[2] - x[1]))
            for i, (label, _start, _end, _ratio) in enumerate(sorted_matches):
                on_match = self._callbacks.get(label)
                if on_match:
                    on_match(self, doc, i, sorted_matches)
            return sorted_matches
        else:
            return []

    def __contains__(self: _PhraseMatcher, label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: _PhraseMatcher) -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    @property
    def labels(self: _PhraseMatcher) -> tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique string labels as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp("Kerouac")])
            >>> matcher.labels
            ('AUTHOR',)
        """
        return tuple(self._patterns.keys())

    @property
    def patterns(self: _PhraseMatcher) -> list[dict[str, Any]]:
        """Get all patterns and kwargs that were added to the matcher.

        Returns:
            The original patterns and kwargs,
            one dictionary for each combination.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp("Kerouac")],
                [{"ignore_case": False}])
            >>> matcher.patterns == [
                {
                    "label": "AUTHOR",
                    "pattern": "Kerouac",
                    "type": "_phrase",
                    "kwargs": {"ignore_case": False}
                    },
                    ]
            True
        """
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(patterns["patterns"], patterns["kwargs"]):
                p = {"label": label, "pattern": pattern.text, "type": self.type}
                if kwargs:
                    p["kwargs"] = kwargs
                all_patterns.append(p)
        return all_patterns

    @property
    def vocab(self: _PhraseMatcher) -> Vocab:
        """Returns the spaCy `Vocab` object utilized."""
        return self._searcher.vocab

    def add(
        self: _PhraseMatcher,
        label: str,
        patterns: list[Doc],
        kwargs: Optional[list[dict[str, Any]]] = None,
        on_match: Optional[
            Callable[[Any, Doc, int, list[tuple[str, int, int, int]]], None]
        ] = None,  # Any type due to inheritence typing issue.
    ) -> None:
        """Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of `Doc` object and if kwargs is not None,
        kwargs must be a list of dictionaries.

        Args:
            label: Name of the rule added to the matcher.
            patterns: `Doc` objects that will be matched
                against the `Doc` object the matcher is called on.
            kwargs: Optional arguments to modify the behavior of the matching.
                Apply to inherited multi_match method.
                See `_PhraseSearcher` documentation for kwarg details.
                Default is `None`.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Raises:
            TypeError: Patterns must be a list of `Doc` objects.
            TypeError: If kwargs is not an iterable dictionaries.

        Warnings:
            KwargsWarning:
                If there are more patterns than kwargs
                default matching settings will be used
                for extra patterns.
            KwargsWarning:
                If there are more kwargs dicts than patterns,
                the extra kwargs will be ignored.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp("mooo")])
            >>> "SOUND" in matcher
            True
        """
        if not isinstance(patterns, list):
            raise TypeError("Patterns must be a list of `Doc objects.")
        if kwargs is None:
            kwargs = [{} for _ in patterns]
        elif len(kwargs) < len(patterns):
            warnings.warn(
                """There are more patterns then there are kwargs.\n
                Patterns not matched to a kwarg dict will have default settings.""",
                KwargsWarning,
            )
            kwargs.extend([{} for _ in range(len(patterns) - len(kwargs))])
        elif len(kwargs) > len(patterns):
            warnings.warn(
                """There are more kwargs dicts than patterns.\n
                The extra kwargs will be ignored.""",
                KwargsWarning,
            )
        for pattern, kwarg in zip(patterns, kwargs):
            if isinstance(pattern, Doc):
                self._patterns[label]["patterns"].append(pattern)
            else:
                raise TypeError("Patterns must be a list of `Doc` objects.")
            if isinstance(kwarg, dict):
                self._patterns[label]["kwargs"].append(kwarg)
            else:
                raise TypeError("Kwargs must be a list of dictionaries.")
        self._callbacks[label] = on_match

    def remove(self: _PhraseMatcher, label: str) -> None:
        """Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Raises:
            ValueError: If label does not exist in the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp("mooo")])
            >>> matcher.remove("SOUND")
            >>> "SOUND" in matcher
            False
        """
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: {label} does not exist within the matcher rules."
            )

    def pipe(
        self: _PhraseMatcher,
        stream: Iterable[Doc],
        batch_size: int = 1000,
        return_matches: bool = False,
        as_tuples: bool = False,
    ) -> Generator[Any, None, None]:
        """Match a stream of `Doc` objects, yielding them in turn.

        Args:
            stream: A stream of `Doc` objects.
            batch_size: Number of documents to accumulate into a working set.
                Default is `1000`.
            return_matches: Yield the match lists along with the docs,
                making results (doc, matches) tuples. Default is `False`.
            as_tuples: Interpret the input stream as (doc, context) tuples,
                and yield (result, context) tuples out.
                If both return_matches and as_tuples are `True`,
                the output will be a sequence of ((doc, matches), context) tuples.
                Default is `False`.

        Yields:
            `Doc` objects, in order.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import _PhraseMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = _PhraseMatcher(nlp.vocab)
            >>> doc_stream = (
                    nlp("test doc1: Korvold"),
                    nlp("test doc2: Prossh"),
                )
            >>> matcher.add("DRAGON", [nlp("Korvold"), nlp("Prossh")])
            >>> output = matcher.pipe(doc_stream, return_matches=True)
            >>> [entry[1] for entry in output]
            [[('DRAGON', 3, 4, 100)], [('DRAGON', 3, 4, 100)]]
        """
        if as_tuples:
            for doc, context in stream:
                matches = self(doc)
                if return_matches:
                    yield ((doc, matches), context)
                else:
                    yield (doc, context)
        else:
            for doc in stream:
                matches = self(doc)
                if return_matches:
                    yield (doc, matches)
                else:
                    yield doc



"""Module for TokenSearcher: flexible token searching in spaCy `Doc` objects."""
from __future__ import annotations

from typing import Any, Optional, Union

import regex
from spacy.tokens import Doc, Token
from spacy.vocab import Vocab

from ..process import FuzzyFuncs, n_wise


class TokenSearcher:
    """Class for flexbile token searching in spaCy `Doc` objects.

    Uses individual (and extended) spaCy token matching patterns to find
    match candidates. Candidates are used to generate new patterns to add
    to a spaCy `Matcher`.

    "FUZZY" and "FREGEX" are the two additional spaCy token pattern options.

    For example:
        {"TEXT": {"FREGEX": "(database){e<=1}"}},
        {"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "quick_lev"}}

    Make sure to use uppercase dictionary keys in patterns.

    Attributes:
        vocab (Vocab): The shared vocabulary.
            Included for consistency and potential future-state.
        _fuzzy_funcs (FuzzyFuncs):
            Container class housing fuzzy matching functions.
            Functions are accessible via the classes `get()` method
            by their given key name. The following rapidfuzz matching
            functions with default settings are available:
            "simple" = `ratio`
            "quick" = `QRatio`
            "quick_lev" = `quick_lev_ratio`
    """

    def __init__(self: TokenSearcher, vocab: Vocab) -> None:
        """Initializes a token searcher.

        Args:
            vocab: A spaCy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are mostly pure-Python
                currently and do not share vocabulary
                with spaCy pipelines.
        """
        self.vocab = vocab
        self._fuzzy_funcs: FuzzyFuncs = FuzzyFuncs(match_type="token")

    def fuzzy_compare(
        self: TokenSearcher,
        a: str,
        b: str,
        ignore_case: bool = True,
        fuzzy_func: str = "simple",
    ) -> int:
        """Peforms fuzzy matching between two strings.

        Applies the given fuzzy matching algorithm (fuzzy_func)
        to two strings and returns the resulting fuzzy ratio.

        Args:
            a: First string for comparison.
            b: Second string for comparison.
            ignore_case: Whether to lower-case a and b
                before comparison or not. Default is `True`.
            fuzzy_func: Key name of fuzzy matching function to use.
                The following rapidfuzz matching functions with default
                settings are available:
                "simple" = `ratio`
                "quick" = `QRatio`
                "quick_lev" = `quick_lev_ratio`
                Default is `"simple"`.

        Returns:
            The fuzzy ratio between a and b.

        Example:
            >>> import spacy
            >>> from spaczz.search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp.vocab)
            >>> searcher.fuzzy_compare("spaczz", "spacy")
            73
        """
        if ignore_case:
            a = a.lower()
            b = b.lower()
        return round(self._fuzzy_funcs.get(fuzzy_func)(a, b))

    def match(
        self: TokenSearcher,
        doc: Doc,
        pattern: list[dict[str, Any]],
        min_r: int = 75,
        fuzzy_func: str = "simple",
    ) -> list[list[Optional[tuple[str, str]]]]:
        """Finds potential token pattern matches in a `Doc` object.

        Make sure to use uppercase dictionary keys in patterns.

        Args:
            doc: `Doc` object to search over.
            pattern: Individual spaCy token pattern.
            min_r: Minimum match ratio required for fuzzy matching.
                Can be overwritten with token pattern options.
                Default is `75`.
            fuzzy_func: Fuzzy matching function to use.
                Can be overwritten with token pattern options.
                Default is `simple`.

        Returns:
            A list of lists with each inner list representing a potential match.
            The inner lists will be populated with key, value tuples of token matches
            and `None` for placeholder tokens representing non-fuzzy tokens.

        Raises:
            TypeError: doc must be a `Doc` object.
            TypeError: pattern must be a `Sequence`.
            ValueError: pattern cannot have zero tokens.

        Example:
            >>> import spacy
            >>> from spaczz.search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp)
            >>> doc = nlp("I was prescribed zithramax and advar")
            >>> pattern = [
                {"TEXT": {"FUZZY": "zithromax"}},
                {"POS": "CCONJ"},
                {"TEXT": {"FREGEX": "(advair){e<=1}"}}
                ]
            >>> searcher.match(doc, pattern)
            [[('TEXT', 'zithramax'), None, ('TEXT', 'advar')]]
        """
        if not isinstance(doc, Doc):
            raise TypeError("doc must be a Doc object.")
        if not isinstance(pattern, list):
            raise TypeError(
                "pattern must be a list", "Make sure pattern is wrapped in a list.",
            )
        if len(pattern) == 0:
            raise ValueError("pattern cannot have zero tokens.")
        matches = []
        for seq in n_wise(doc, len(pattern)):
            seq_matches = self._iter_pattern(seq, pattern, min_r, fuzzy_func)
            if seq_matches:
                matches.append(seq_matches)
        if matches:
            filtered_matches = [
                i for n, i in enumerate(matches) if i not in matches[:n]
            ]
            return filtered_matches
        else:
            return matches

    @staticmethod
    def regex_compare(text: str, pattern: str, ignore_case: bool = False) -> bool:
        """Performs fuzzy-regex supporting regex matching between two strings.

        Args:
            text: The string to match against.
            pattern: The regex pattern string.
            ignore_case: Whether to lower-case text
                before comparison or not. Default is `False`.

        Returns:
            `True` if match, `False` if not.

        Example:
            >>> import spacy
            >>> from spaczz.search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp)
            >>> searcher.regex_compare("sequel", "(sql){i<=3}")
            True
        """
        if ignore_case:
            text = text.lower()
        if regex.match(pattern, text):
            return True
        else:
            return False

    def _iter_pattern(
        self: TokenSearcher,
        seq: tuple[Token, ...],
        pattern: list[dict[str, Any]],
        min_r: int,
        fuzzy_func: str,
    ) -> list[Optional[tuple[str, str]]]:
        """Evaluates each token in a pattern against a doc token sequence."""
        seq_matches: list[Optional[tuple[str, str]]] = []
        for i, token in enumerate(pattern):
            pattern_dict, case, case_bool = self._parse_case(token)
            if isinstance(pattern_dict, dict):
                pattern_text, pattern_type = self._parse_type(pattern_dict)
                if pattern_text and pattern_type == "FUZZY":
                    if self.fuzzy_compare(
                        seq[i].text,
                        pattern_text,
                        case_bool,
                        pattern_dict.get("FUZZY_FUNC", fuzzy_func),
                    ) >= pattern_dict.get("MIN_R", min_r):
                        seq_matches.append((case, seq[i].text))
                    else:
                        return []
                elif pattern_text and pattern_type == "FREGEX":
                    if self.regex_compare(seq[i].text, pattern_text, case_bool):
                        seq_matches.append((case, seq[i].text))
                    else:
                        return []
                else:
                    seq_matches.append(None)
            else:
                seq_matches.append(None)
        return seq_matches

    @staticmethod
    def _parse_case(token: dict[str, Any]) -> tuple[Union[str, dict, None], str, bool]:
        """Parses the case of a token pattern."""
        if token.get("TEXT"):
            return token.get("TEXT"), "TEXT", False
        else:
            return token.get("LOWER"), "LOWER", True

    @staticmethod
    def _parse_type(pattern_dict: dict[str, Any]) -> tuple[str, str]:
        """Parses the type of a token pattern."""
        fuzzy_text = pattern_dict.get("FUZZY")
        regex_text = pattern_dict.get("FREGEX")
        if isinstance(fuzzy_text, str):
            return fuzzy_text, "FUZZY"
        elif isinstance(regex_text, str):
            return regex_text, "FREGEX"
        else:
            return "", ""



"""Module for RegexMatcher with an API semi-analogous to spaCy's PhraseMatcher."""
from __future__ import annotations

from collections import defaultdict
from typing import (
    Any,
    Callable,
    DefaultDict,
    Generator,
    Iterable,
    Optional,
    Union,
)
import warnings

from spacy.tokens import Doc
from spacy.vocab import Vocab

from ..exceptions import KwargsWarning
from ..regex import RegexConfig
from ..search import RegexSearcher


class RegexMatcher:
    """spaCy-like matcher for finding multi-token regex matches in `Doc` objects.

    Matches added patterns against the `Doc` object it is called on.
    Accepts labeled regex patterns in the form of strings.

    Attributes:
        defaults: Keyword arguments to be used as default matching settings.
            See `RegexSearcher` documentation for details.
        name: Class attribute - the name of the matcher.
        type: The kind of matcher object.
        _callbacks:
            On match functions to modify `Doc` objects passed to the matcher.
            Can make use of the matches identified.
        _patterns:
            Patterns added to the matcher. Contains patterns
            and kwargs that should be used during matching
            for each labels added.
    """

    name = "regex_matcher"

    def __init__(
        self: RegexMatcher,
        vocab: Vocab,
        config: Union[str, RegexConfig] = "default",
        **defaults: Any,
    ) -> None:
        """Initializes the regex matcher with the given config and defaults.

        Args:
            vocab: A spacy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are currently pure
                Python and do not share vocabulary
                with spacy pipelines.
            config: Provides predefind regex patterns and flags.
                Uses the default config if "default", an empty config if "empty",
                or a custom config by passing a `RegexConfig` object.
                Default is "default".
            **defaults: Keyword arguments that will
                be used as default matching settings.
                These arguments will become the new defaults for matching.
                See `RegexSearcher` documentation for details.
        """
        self.defaults = defaults
        self.type = "regex"
        self._callbacks: dict[
            str,
            Optional[
                Callable[
                    [
                        RegexMatcher,
                        Doc,
                        int,
                        list[tuple[str, int, int, tuple[int, int, int]]],
                    ],
                    None,
                ],
            ],
        ] = {}
        self._patterns: DefaultDict[str, DefaultDict[str, Any]] = defaultdict(
            lambda: defaultdict(list)
        )
        self._searcher = RegexSearcher(vocab=vocab, config=config)

    def __call__(
        self: RegexMatcher, doc: Doc
    ) -> list[tuple[str, int, int, tuple[int, int, int]]]:
        r"""Find all sequences matching the supplied patterns in the doc.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of (key, start, end, fuzzy change count) tuples,
            describing the matches.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> doc = nlp.make_doc("I live in the united states, or the US")
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> matcher(doc)
            [('GPE', 4, 6, (0, 0, 0)), ('GPE', 9, 10, (0, 0, 0))]
        """
        matches = set()
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(patterns["patterns"], patterns["kwargs"]):
                if not kwargs:
                    kwargs = self.defaults
                matches_wo_label = self._searcher.match(doc, pattern, **kwargs)
                if matches_wo_label:
                    matches_w_label = [
                        (label,) + match_wo_label for match_wo_label in matches_wo_label
                    ]
                    for match in matches_w_label:
                        matches.add(match)
        if matches:
            sorted_matches = sorted(matches, key=lambda x: (x[1], -x[2] - x[1]))
            for i, (label, _start, _end, _subs) in enumerate(sorted_matches):
                on_match = self._callbacks.get(label)
                if on_match:
                    on_match(self, doc, i, sorted_matches)
            return sorted_matches
        else:
            return []

    def __contains__(self: RegexMatcher, label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: RegexMatcher) -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    @property
    def labels(self: RegexMatcher) -> tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique string labels as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("ZIP", ["zip_codes"], [{"predef": True}])
            >>> matcher.labels
            ('ZIP',)
        """
        return tuple(self._patterns.keys())

    @property
    def patterns(self: RegexMatcher) -> list[dict[str, Any]]:
        """Get all patterns and kwargs that were added to the matcher.

        Returns:
            The original patterns and kwargs,
            one dictionary for each combination.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("ZIP", ["zip_codes"], [{"predef": True}])
            >>> matcher.patterns == [
                {
                    "label": "ZIP",
                    "pattern": "zip_codes",
                    "type": "regex",
                    "kwargs": {"predef": True},
                    }
                    ]
            True
        """
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(patterns["patterns"], patterns["kwargs"]):
                p = {"label": label, "pattern": pattern, "type": "regex"}
                if kwargs:
                    p["kwargs"] = kwargs
                all_patterns.append(p)
        return all_patterns

    @property
    def vocab(self: RegexMatcher) -> Vocab:
        """Returns the spaCy `Vocab` object utilized."""
        return self._searcher.vocab

    def add(
        self: RegexMatcher,
        label: str,
        patterns: list[str],
        kwargs: Optional[list[dict[str, Any]]] = None,
        on_match: Optional[
            Callable[
                [
                    RegexMatcher,
                    Doc,
                    int,
                    list[tuple[str, int, int, tuple[int, int, int]]],
                ],
                None,
            ]
        ] = None,
    ) -> None:
        r"""Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of strings and if kwargs is not `None`,
        kwargs must be a list of dictionaries.

        To utilize regex flags, use inline flags.

        Args:
            label: Name of the rule added to the matcher.
            patterns: Strings that will be matched against
                the Doc object the matcher is called on.
            kwargs: Optional arguments to modify the behavior of the regex matching.
                Apply to inherited multi_match method.
                Default is `None`.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Raises:
            TypeError: If patterns is not a non-string iterable of strings.
            TypeError: If kwargs is not a iterable of dictionaries.

        Warnings:
            KwargsWarning:
                If there are more patterns than kwargs
                default regex matching settings will be used
                for extra patterns.
            KwargsWarning:
                If there are more kwargs dictionaries than patterns,
                the extra kwargs will be ignored.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> "GPE" in matcher
            True
        """
        if kwargs is None:
            kwargs = [{} for _ in patterns]
        elif len(kwargs) < len(patterns):
            warnings.warn(
                """There are more patterns then there are kwargs.\n
                    Patterns not matched to a kwarg dict will have default settings.""",
                KwargsWarning,
            )
            kwargs.extend([{} for _ in range(len(patterns) - len(kwargs))])
        elif len(kwargs) > len(patterns):
            warnings.warn(
                """There are more kwargs dicts than patterns.\n
                    The extra kwargs will be ignored.""",
                KwargsWarning,
            )
        if isinstance(patterns, str):
            raise TypeError("Patterns must be a non-string iterable of strings.")
        for pattern, kwarg in zip(patterns, kwargs):
            if isinstance(pattern, str):
                self._patterns[label]["patterns"].append(pattern)
            else:
                raise TypeError("Patterns must be a non-string iterable of strings.")
            if isinstance(kwarg, dict):
                self._patterns[label]["kwargs"].append(kwarg)
            else:
                raise TypeError("Kwargs must be an iterable of dictionaries.")
        self._callbacks[label] = on_match

    def remove(self: RegexMatcher, label: str) -> None:
        r"""Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Raises:
            ValueError: If label does not exist in the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> matcher.remove("GPE")
            >>> "GPE" in matcher
            False
        """
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: {label} does not exist within the matcher rules."
            )

    def pipe(
        self: RegexMatcher,
        stream: Iterable[Doc],
        batch_size: int = 1000,
        return_matches: bool = False,
        as_tuples: bool = False,
    ) -> Generator[Any, None, None]:
        r"""Match a stream of `Doc` objects, yielding them in turn.

        Args:
            stream: A stream of `Doc` objects.
            batch_size: Number of documents to accumulate into a working set.
                Default is `1000`.
            return_matches: Yield the match lists along with the docs,
                making results (doc, matches) tuples. Default is `False`.
            as_tuples: Interpret the input stream as (doc, context) tuples,
                and yield (result, context) tuples out.
                If both return_matches and as_tuples are `True`,
                the output will be a sequence of ((doc, matches), context) tuples.
                Default is `False`.

        Yields:
            Doc objects, in order.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> doc_stream = (
                    nlp.make_doc("test doc1: United States"),
                    nlp.make_doc("test doc2: US"),
                )
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> output = matcher.pipe(doc_stream, return_matches=True)
            >>> [entry[1] for entry in output]
            [[('GPE', 3, 5, (0, 0, 0))], [('GPE', 3, 4, (0, 0, 0))]]
        """
        if as_tuples:
            for doc, context in stream:
                matches = self(doc)
                if return_matches:
                    yield ((doc, matches), context)
                else:
                    yield (doc, context)
        else:
            for doc in stream:
                matches = self(doc)
                if return_matches:
                    yield (doc, matches)
                else:
                    yield doc



"""Module for _PhraseSearcher: flexible phrase searching in spaCy `Doc` objects."""
from __future__ import annotations

from itertools import chain
from typing import Any, Union
import warnings

from spacy.tokens import Doc, Span, Token
from spacy.vocab import Vocab

from ..exceptions import FlexWarning, RatioWarning


class _PhraseSearcher:
    """Base class for flexible phrase searching in spaCy `Doc` objects.

    Phrase matching is done on the token level.

    Not intended for use as-is. All methods and attributes except
    the `.compare()` method are shared with the `FuzzySearcher` and
    `SimilaritySearcher`.

    Attributes:
        vocab (Vocab): The shared vocabulary.
            Included for consistency and potential future-state.
    """

    def __init__(self: _PhraseSearcher, vocab: Vocab) -> None:
        """Initializes a base phrase searcher.

        Args:
            vocab: A spaCy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are mostly pure-Python
                currently and do not share vocabulary
                with spaCy pipelines.
        """
        self.vocab = vocab

    def compare(
        self: _PhraseSearcher,
        a: Union[Doc, Span, Token],
        b: Union[Doc, Span, Token],
        ignore_case: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> int:
        """Base method for comparing two spaCy containers.

        Tests the equality between the unicode string representation
        of each container (`Doc`, `Span`, `Token`) and returns
        an integer boolean representation as 100 or 0.

        Will be overwritten in child classes.

        Args:
            a: First container for comparison.
            b: Second container for comparison.
            ignore_case: Whether to lower-case a and b
                before comparison or not. Default is `True`.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            100 if equal, 0 if not.
        """
        if ignore_case:
            a_text = a.text.lower()
            b_text = b.text.lower()
        else:
            a_text = a.text
            b_text = b.text
        if a_text == b_text:
            return 100
        else:
            return 0

    def match(
        self: _PhraseSearcher,
        doc: Doc,
        query: Doc,
        flex: Union[str, int] = "default",
        min_r1: int = 50,
        min_r2: int = 75,
        thresh: int = 100,
        *args: Any,
        **kwargs: Any,
    ) -> list[tuple[int, int, int]]:
        """Returns phrase matches in a `Doc` object.

        Finds phrase matches in the doc based on the query,
        assuming the minimum match ratios (min_r1 and min_r2) are met.
        Matches will be sorted by descending matching score,
        then ascending start index.

        Args:
            doc: `Doc` object to search over.
            query: `Doc` object to match against doc.
            flex: Number of tokens to move match match boundaries
                left and right during optimization.
                Can be an integer value with a max of `len(query)`
                and a min of `0` (will warn and change if higher or lower),
                or the strings "max", "min", or "default".
                Default is `"default"`: `len(query) // 2`.
            min_r1: Minimum match ratio required for
                selection during the intial search over doc.
                If `flex == 0`, `min_r1` will be overwritten by `min_r2`.
                If `flex > 0`, `min_r1` must be lower than `min_r2`
                and "low" in general because match boundaries are
                not flexed initially.
                Default is `50`.
            min_r2: Minimum match ratio required for
                selection during match optimization.
                Needs to be higher than `min_r1` and "high" in general
                to ensure only quality matches are returned.
                Default is `75`.
            thresh: If this ratio is exceeded in initial scan,
                and `flex > 0`, no optimization will be attempted.
                If `flex == 0`, `thresh` has no effect.
                Default is `100`.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            A list of tuples of match span start indices,
            end indices, and match ratios.

        Raises:
            TypeError: doc must be a `Doc` object.
            TypeError: query must be a `Doc` object.
        """
        if not isinstance(doc, Doc):
            raise TypeError("doc must be a Doc object.")
        if not isinstance(query, Doc):
            raise TypeError("query must be a Doc object.")
        flex = self._calc_flex(query, flex)
        min_r1, min_r2, thresh = self._check_ratios(min_r1, min_r2, thresh, flex)
        match_values = self._scan(doc, query, min_r1, *args, **kwargs)
        if match_values:
            positions = list(match_values.keys())
            matches_w_nones = [
                self._optimize(
                    doc,
                    query,
                    match_values,
                    pos,
                    flex,
                    min_r2,
                    thresh,
                    *args,
                    **kwargs,
                )
                for pos in positions
            ]
            matches = [match for match in matches_w_nones if match]
            if matches:
                sorted_matches = sorted(matches, key=lambda x: (-x[2], x[0]))
                filtered_matches = self._filter_overlapping_matches(sorted_matches)
                return filtered_matches
            else:
                return []
        else:
            return []

    def _optimize(
        self: _PhraseSearcher,
        doc: Doc,
        query: Doc,
        match_values: dict[int, int],
        pos: int,
        flex: int,
        min_r2: int,
        thresh: int,
        *args: Any,
        **kwargs: Any,
    ) -> Union[tuple[int, int, int], None]:
        """Optimizes a potential match by flexing match span boundaries.

        For a span match from _scan that has match ratio
        greater than or equal to min_r1 the span boundaries
        will be extended both left and right by flex number
        of tokens and matched back to the original query.
        The optimal start and end index are then returned
        along with the span's match ratio.

        Args:
            doc: `Doc` object being searched over.
            query: `Doc` object to match against doc.
            match_values: Dictionary of initial match spans
                start indices and match ratios.
            pos: Start index of span being optimized.
            flex: Number of tokens to move match span boundaries
                left and right during match optimization.
            min_r2: Minimum match ratio required
                to pass optimization. This should be high enough
                to only return quality matches.
            thresh: If this ratio is exceeded in initial scan,
                no optimization will be attempted.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            A tuple of left boundary index,
            right boudary index, and match ratio
            or `None`.
        """
        p_l, bp_l = [pos] * 2
        p_r, bp_r = [pos + len(query)] * 2
        r = match_values[pos]
        if flex and not r >= thresh:
            optim_r = r
            for f in range(1, flex + 1):
                if p_l - f >= 0:
                    new_r = self.compare(query, doc[p_l - f : p_r], *args, **kwargs)
                    if new_r > optim_r:
                        optim_r = new_r
                        bp_l = p_l - f
                if p_l + f < min(p_r, bp_r):
                    new_r = self.compare(query, doc[p_l + f : p_r], *args, **kwargs)
                    if new_r > optim_r:
                        optim_r = new_r
                        bp_l = p_l + f
                if p_r - f > max(p_l, bp_l):
                    new_r = self.compare(query, doc[p_l : p_r - f], *args, **kwargs)
                    if new_r > optim_r:
                        optim_r = new_r
                        bp_r = p_r - f
                if p_r + f <= len(doc):
                    new_r = self.compare(query, doc[p_l : p_r + f], *args, **kwargs)
                    if new_r > optim_r:
                        optim_r = new_r
                        bp_r = p_r + f
                if optim_r <= r:
                    break
                else:
                    r = optim_r
        if r >= min_r2:
            return (bp_l, bp_r, r)
        else:
            return None

    def _scan(
        self: _PhraseSearcher,
        doc: Doc,
        query: Doc,
        min_r1: int,
        *args: Any,
        **kwargs: Any,
    ) -> Union[dict[int, int], None]:
        """Returns a dictionary of potential match start indices and match ratios.

        Iterates through the doc by spans of query length,
        and matches each span against query.

        If a span's match ratio is greater than or equal to the
        min_r1 it is added to a dict with it's start index
        as the key and it's ratio as the value.

        Args:
            doc: `Doc` object to search over.
            query: `Doc` object to match against doc.
            min_r1: Minimum match ratio required for
                selection during the intial search over doc.
                This should be lower than min_r2 and "low" in general
                because match span boundaries are not flexed here.
                0 means all spans of query length in doc will
                have their boundaries flexed and will be recompared
                during match optimization.
                Lower min_r1 will result in more fine-grained matching
                but will run slower.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            A dictionary of start index, match ratio pairs or None.
        """
        if not len(query):
            return None
        match_values: dict[int, int] = dict()
        i = 0
        while i + len(query) <= len(doc):
            match = self.compare(query, doc[i : i + len(query)], *args, **kwargs)
            if match >= min_r1:
                match_values[i] = match
            i += 1
        if match_values:
            return match_values
        else:
            return None

    @staticmethod
    def _calc_flex(query: Doc, flex: Union[str, int]) -> int:
        """Returns flex value based on initial value and query.

        By default flex is set to `len(query) // 2`.

        If flex is an integer value greater than `len(query)`,
        flex will be set to that value instead.

        If flex is an integer value less than 0,
        flex will be set to 0 instead.

        Args:
            query: The `Doc` object to match with.
            flex: Either `"default"`: `len(query) // 2`,
                `"max"`: `len(query)`,
                `"min"`: `0`,
                or an integer value.

        Returns:
            The new flex value.

        Raises:
            TypeError: If flex is not "default", "max", "min", or an int.

        Warnings:
            FlexWarning:
                If flex is > `len(query)`.
            FlexWarning:
                If flex is < 0.

        Example:
            >>> import spacy
            >>> from spaczz.search import _PhraseSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = _PhraseSearcher(nlp.vocab)
            >>> query = nlp("Test query")
            >>> searcher._calc_flex(query, "default")
            1
        """
        if flex == "default":
            flex = len(query) // 2
        if flex == "max":
            flex = len(query)
        if flex == "min":
            flex = 0
        elif isinstance(flex, int):
            if flex > len(query):
                warnings.warn(
                    f"""`flex` of size {flex} is greater than `len(query)`.
                        Setting to that max value instead.""",
                    FlexWarning,
                )
                flex = len(query)
            if flex < 0:
                warnings.warn(
                    """`flex` values less than `0` are not allowed.
                    Setting to the min, `0`, instead.""",
                    FlexWarning,
                )
                flex = 0
        else:
            raise TypeError(
                (
                    "Flex must either be the strings 'default',",
                    "`max`, or `min`, or an integer.",
                )
            )
        return flex

    @staticmethod
    def _check_ratios(
        min_r1: int, min_r2: int, thresh: int, flex: int
    ) -> tuple[int, int, int]:
        """Ensures ratios are not set to illegal values."""
        if flex:
            if min_r1 > min_r2:
                warnings.warn(
                    """`min_r1` > `min_r2`,
                setting `min_r1` equal to `min_r2`""",
                    RatioWarning,
                )
                min_r1 = min_r2
            if thresh < min_r2:
                warnings.warn(
                    """`thresh` < `min_r2`,
                setting `thresh` equal to `min_r2`""",
                    RatioWarning,
                )
                thresh = min_r2
        else:
            min_r1 = min_r2
        return min_r1, min_r2, thresh

    @staticmethod
    def _filter_overlapping_matches(
        matches: list[tuple[int, int, int]]
    ) -> list[tuple[int, int, int]]:
        """Prevents multiple match spans from overlapping.

        Expects matches to be pre-sorted by descending ratio
        then ascending start index.
        If more than one match span includes the same tokens
        the first of these match spans in matches is kept.

        Args:
            matches: List of match span tuples
                (start_index, end_index, fuzzy ratio).

        Returns:
            The filtered list of match span tuples.

        Example:
            >>> import spacy
            >>> from spaczz.search import _PhraseSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = _PhraseSearcher(nlp.vocab)
            >>> matches = [(1, 3, 80), (1, 2, 70)]
            >>> searcher._filter_overlapping_matches(matches)
            [(1, 3, 80)]
        """
        filtered_matches: list[tuple[int, int, int]] = []
        for match in matches:
            if not set(range(match[0], match[1])).intersection(
                chain(*[set(range(n[0], n[1])) for n in filtered_matches])
            ):
                filtered_matches.append(match)
        return filtered_matches



"""Module for matchers."""
from ._phrasematcher import _PhraseMatcher
from .fuzzymatcher import FuzzyMatcher
from .regexmatcher import RegexMatcher
from .similaritymatcher import SimilarityMatcher
from .tokenmatcher import TokenMatcher

__all__ = [
    "_PhraseMatcher",
    "FuzzyMatcher",
    "RegexMatcher",
    "SimilarityMatcher",
    "TokenMatcher",
]
